from ignite.contrib.metrics.average_precision import AveragePrecision
from ignite.contrib.metrics.roc_auc import ROC_AUC
import ignite.contrib.metrics.regression
from ignite.contrib.metrics.gpu_info import GpuInfo
